﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using CSScriptIntellisense;
using Res=CSScriptIntellisense.Images;
using ICSharpCode.NRefactory.Completion;
using ICSharpCode.NRefactory.TypeSystem;

namespace CSScriptIntellisense
{
    public partial class AutocompleteForm : Form
    {
        Action<ICompletionData> OnAutocompletionAccepted;

        public AutocompleteForm()
        {
            InitializeComponent();

            OnAutocompletionAccepted = x => { };
        }

        string initialPartialName = null;

        public AutocompleteForm(Action<ICompletionData> onAutocompletionAccepted, IEnumerable<ICompletionData> items, string initialSuggestionHint)
        {
            InitializeComponent();

            OnAutocompletionAccepted = onAutocompletionAccepted;

            rawItems = items;

            initialPartialName = initialSuggestionHint;
        }

        int itemHeight;

        public void FilterFor(string partialName)
        {
            listBox1.Items.Clear();

            IEnumerable<ICompletionData> items = ProcessSuggestionHint(partialName, rawItems);

            listBox1.Items.AddRange(items.ToArray());
            listBox1.SelectedItem = items.FirstOrDefault();

            var g = listBox1.CreateGraphics();
            var wideItem = items.Select(x => (int)g.MeasureString(x.DisplayText, listBox1.Font).Width).Max(x => x);
            this.Width = Math.Min(wideItem + 40, 250);//40 = 20 for icon on left and 20 for scrollbar on right

            this.Height = ((itemHeight + verticalSpacing) * Math.Min(listBox1.Items.Count, 10)) + 5;
        }

        public IEnumerable<ICompletionData> ProcessSuggestionHint(string partialName, IEnumerable<ICompletionData> inputItems)
        {
            if (string.IsNullOrWhiteSpace(partialName))
            {
                return inputItems;
            }
            else
            {
                //try partial matches (both the text an the pattern)
                var query = from item in inputItems
                            let matchCount = item.DisplayText.MatchingStartChars(partialName, ignoreCase: true)
                            group item by matchCount into g
                            orderby g.Key descending
                            select new { MatchingDegree = g.Key, Items = g };

                if (query.Count() > 0)
                    return query.First().Items;
                else
                    return inputItems;
            }
        }

        int verticalSpacing = 6;

        void listBox1_MeasureItem(object sender, MeasureItemEventArgs e)
        {
            var item = (ICompletionData)listBox1.Items[e.Index];

            string itemString = item.DisplayText;

            SizeF size = e.Graphics.MeasureString(item.DisplayText, listBox1.Font);
            e.ItemHeight = (int)size.Height + verticalSpacing;
            e.ItemWidth = (int)size.Width + 16 + 10 + 10; //ensure enough space for the icon and the scrollbar
            listBox1.HorizontalExtent = e.ItemWidth;
        }

        Dictionary<string, Image> icons = new Dictionary<string, Image>();

        Image GetImageFor(ICompletionData item)
        {
            var itemType = EntityType.None;
            var declarationType = DeclarationType.None;
            bool isExtensionMethod = false;

            if (item is IEntityCompletionData)
            {
                itemType = (item as IEntityCompletionData).Entity.EntityType;
            }

            if (item is CompletionData)
            {
                var data = (item as CompletionData);
                declarationType = data.DeclarationType;
                isExtensionMethod = data.IsExtensionMethod;
            }

            switch (itemType)
            {
                case EntityType.Constructor:
                case EntityType.Destructor:
                    return Res.Images.constructor;
                case EntityType.Method:
                    {
                        if (isExtensionMethod)
                            return Res.Images.extension_method;
                        else
                            return Res.Images.method;
                    }
                case EntityType.Event:
                    return Res.Images._event;
                case EntityType.Field:
                    return Res.Images.field;
                case EntityType.Property:
                    return Res.Images.property;
                default:
                    break;
            }

            switch (declarationType)
            {
                case DeclarationType.None:
                    break;
                case DeclarationType.Namespace:
                    return Res.Images._namespace;
                case DeclarationType.Type:
                    return Res.Images.constructor;
                case DeclarationType.Variable:
                case DeclarationType.Parameter:
                    return Res.Images.field;
                case DeclarationType.Event:
                    return Res.Images._event;
                case DeclarationType.Unresolved:
                    return Res.Images.unresolved;
                default:
                    break;
            }

            return null;
        }

        void listBox1_DrawItem(object sender, DrawItemEventArgs e)
        {
            if (e.Index == -1)
                return;

            var item = (ICompletionData)listBox1.Items[e.Index];

            e.DrawBackground();

            Brush brush = Brushes.Black;

            if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
                brush = Brushes.White;

            Rectangle rect = e.Bounds;
            rect.Offset(16, 3);
            e.Graphics.DrawString(item.DisplayText, e.Font, brush, rect, StringFormat.GenericDefault);

            var image = GetImageFor(item);
            if (image != null)
            {
                Rectangle r = e.Bounds;
                r.Width = 16;
                e.Graphics.FillRectangle(Brushes.White, r);
                e.Graphics.DrawImage(image, e.Bounds.Location);
            }
        }

        IEnumerable<ICompletionData> rawItems;

        private void AutocompleteForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Escape)
                Close();

            if (e.KeyCode == Keys.Enter || (e.KeyCode == Keys.Right && Config.Instance.UseArrowToAccept))
            {
                Close();
                OnAutocompletionAccepted(listBox1.SelectedItem as ICompletionData);
            }
        }

        private void AutocompleteForm_Deactivate(object sender, EventArgs e)
        {
            Close();
        }

        private void listBox1_DoubleClick(object sender, EventArgs e)
        {
            Close();
            OnAutocompletionAccepted(listBox1.SelectedItem as ICompletionData);
        }

        private void AutocompleteForm_Load(object sender, EventArgs e)
        {
            var g = listBox1.CreateGraphics();
            itemHeight = (int)g.MeasureString("T", listBox1.Font).Height;

            listBox1.Sorted = false;
            listBox1.DrawMode = DrawMode.OwnerDrawVariable;
            listBox1.DrawItem += listBox1_DrawItem;
            listBox1.MeasureItem += listBox1_MeasureItem;
            listBox1.HorizontalScrollbar = true;

            FilterFor(initialPartialName);
        }

        private void AutocompleteForm_KeyPress(object sender, KeyPressEventArgs e)
        {
        }
    }

    static class Extensions
    {
        public static int MatchingStartChars(this string text, string pattern, bool ignoreCase = false)
        {
            if (string.IsNullOrEmpty(text) || string.IsNullOrEmpty(pattern))
                return 0;

            if (ignoreCase)
            {
                text = text.ToLower();
                pattern = pattern.ToLower();
            }

            for (int i = 0; i < pattern.Length && i < text.Length; i++)
            {
                if (text[i] != pattern[i])
                    return i;
            }
            return Math.Min(pattern.Length, text.Length);
        }
    }
}