using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ICSharpCode.NRefactory.CSharp;
using ICSharpCode.NRefactory.Completion;
using ICSharpCode.NRefactory.TypeSystem;

namespace CSScriptIntellisense
{
    public static class NRefactoryExtensions
    {
        public static int GetArgumentCount(string memberInfo)
        {
            //Method: void Class.Method(string text, Dictionary<int, string> map)
            //Method: void Console.WriteLine(bool value)

            int start = memberInfo.IndexOf('(');
            int end = memberInfo.LastIndexOf(')');
            if (start != -1 && end != -1 && (end - start) != 1)
            {
                char[] array = memberInfo.Substring(start, end-start).ToCharArray();

                int count = 1;
                int level = 0; 
                foreach(char c in array)
                {
                    //exclude all comas in the angle brackets
                    if (c == '<') 
                        level++;
                    else if (c == '>')
                        level--;
                    if (level == 0 && c== ',')
                        count++;
                }
                return count;
            }
            else
            {
                return 0;
            }
        }

        public static string ToTooltip(this ICSharpCode.NRefactory.TypeSystem.IEntity entity)
        {
            try
            {
                var builder = new StringBuilder();
                builder.Append(entity.EntityType);
                builder.Append(": ");

                var property = (entity as IProperty);
                var field = (entity as IField);

                var method = (entity as IMethod);
                if (method != null)
                {
                    builder.Append(method.ReturnType.ReflectionName);
                    builder.Append(' ');
                }
                else if (property != null)
                {
                    if (property.CanGet)
                        builder.Append(property.Getter.ReturnType.ReflectionName);
                    else if (property.CanSet)
                        builder.Append(property.Setter.ReturnType.ReflectionName);

                    builder.Append(' ');
                }
                else if (field != null)
                {
                    builder.Append(field.ReturnType.ReflectionName);
                    builder.Append(' ');
                }

                builder.Append(entity.DeclaringType.ReflectionName);
                builder.Append('.');
                builder.Append(entity.Name);

                if (method != null && method.TypeParameters.Count > 0)
                {
                    builder.Append('`');
                    builder.Append(method.TypeParameters.Count);
                    builder.Append('[');
                    for (int i = 0; i < method.TypeParameters.Count; i++)
                    {
                        if (i > 0)
                            builder.Append(",");
                        builder.Append("[``");
                        builder.Append(i);
                        builder.Append("]");
                    }

                    builder.Append(']');
                }

                if (entity.EntityType == EntityType.Method || entity.EntityType == EntityType.Constructor || entity.EntityType == EntityType.Destructor)
                {
                    builder.Append('(');
                    if (method != null)
                    {
                        for (int i = 0; i < method.Parameters.Count; i++)
                        {
                            if (i > 0)
                                builder.Append(", ");

                            object param = method.Parameters[i];
                            string s = param.ToString();
                            builder.Append(method.Parameters[i].ToDisplayString());
                        }
                    }
                    else
                    {
                        builder.Append("...");
                    }
                    builder.Append(')');
                }
                return builder.ToString().ProcessGenricNotations().ReplaceClrAliaces();
            }
            catch
            {
                return null;
            }
        }

        public static string ToDisplayString(this IParameter parameter)
        {
            var builder = new StringBuilder();
            if (parameter.IsRef)
            {
                builder.Append("ref ");
            }
            if (parameter.IsOut)
            {
                builder.Append("out ");
            }
            if (parameter.IsParams)
            {
                builder.Append("params ");
            }
            builder.Append(parameter.Type.ReflectionName);
            if (parameter.IsOptional)
            {
                builder.Append(" = ");
                if (parameter.ConstantValue != null)
                {
                    builder.Append(parameter.ConstantValue.ToString());
                }
                else
                {
                    builder.Append("null");
                }
            }
            builder.Append(' ');
            builder.Append(parameter.Name);

            return builder.ToString();
        }

        static public string HideKnownNamespaces(this string text, params string[] namespaces)
        {
            if (!string.IsNullOrEmpty(text))
                foreach (string item in namespaces.OrderBy(x => x).Reverse())
                    text = text.Replace(item + ".", "");

            return text;
        }

        static public string ReplaceClrAliaces(this string text)
        {
            if (string.IsNullOrEmpty(text))
                return text;
            else
                return text.Replace("System.Object", "object") //will need more effective replacement implementation
                           .Replace("System.Boolean", "bool")
                           .Replace("System.Byte", "byte")
                           .Replace("System.SByte", "sbyte")
                           .Replace("System.Char", "char")
                           .Replace("System.Decimal", "decimal")
                           .Replace("System.Double", "double")
                           .Replace("System.Single", "float")
                           .Replace("System.Int32", "int")
                           .Replace("System.UInt32", "uint")
                           .Replace("System.Int64", "long")
                           .Replace("System.UInt64", "ulong")
                           .Replace("System.Object", "object")
                           .Replace("System.Int16", "short")
                           .Replace("System.UInt16", "ushort")
                           .Replace("System.String", "string")
                           .Replace("Void", "void");
        }

        public static string ProcessGenricNotations(this string text)
        {
            //It is tempting to compose the "tooltip" string from the CLR property values but
            //it has to be done through string conversion as sometimes the name comes as
            //a string already pre-formatted some where deep in the NRefactory (e.g. parameter.Type.ReflectionName)

            const char voidChar = (char)0;

            char[] data = text.ToCharArray();

            for (int i = 0; i < data.Length; i++)
            {
                if (data[i] == '`')
                {
                    if (data.HasCharAt('`', i + 1)) // "``0"
                    {
                        // convert "``1" -> " T1"
                        data[i++] = voidChar;
                        data[i] = 'T';
                        if (data.HasCharAt('0', i + 1))
                        {
                            data[++i] = voidChar; //convert "``0" -> " T "
                        }
                    }
                    else // "`1[...]" but not "``0"
                    {
                        int openBracket = data.FindChar('[', i);
                        if (openBracket != -1 && data.AreNumbers(i + 1, openBracket - 1))
                        {
                            int closingBracket = data.FindMatchingBracket(openBracket);
                            if (closingBracket != -1)
                            {
                                //convert "`1[...]" -> "  <...>"
                                data[i++] = voidChar;
                                data[i++] = voidChar;
                                data[i] = '<';
                                data[closingBracket] = '>';
                            }
                        }
                    }
                }
                else if (data[i] == '[')
                {
                    int closingBracket = data.FindMatchingBracket(i);
                    {
                        if (closingBracket != -1)
                        {
                            if (closingBracket == i + 1)  // "[]"
                            {
                                continue;
                            }
                            else
                            {
                                //convert "[``0]" -> " ``0 "
                                //convert "[bool]" -> " bool "
                                data[i] = voidChar;
                                data[closingBracket] = voidChar;
                            }
                        }
                    }
                }
            }

            char[] processedData = data.Where(c => c != voidChar).ToArray();

            return new string(processedData);
        }

        public static int FindChar(this char[] array, char c, int index)
        {
            if (index < array.Length)
                for (int i = index; i < array.Length; i++)
                    if (array[i] == c)
                        return i;
            return -1;
        }

        public static bool HasCharAt(this char[] array, char c, int index)
        {
            if (index < array.Length)
                return array[index] == c;
            return false;
        }

        public static bool AreNumbers(this char[] array, int start, int end)
        {
            for (int i = start; i <= end; i++)
                if (!Char.IsNumber(array[i]))
                    return false;
            return true;
        }

        public static int FindMatchingBracket(this char[] array, int start)
        {
            int level = 1; //it is opened
            for (int i = start + 1; i < array.Length; i++)
            {
                if (array[i] == '[')
                    level++;
                else if (array[i] == ']')
                    level--;
                if (level == 0)
                    return i;
            }
            return -1;
        }

        public static bool AreBracketsClosed(this string text, char openBracket = '(', char closeBracket = ')')
        {
            int level = 0; 
            for (int i = 0; i < text.Length; i++)
            {
                if (text[i] == openBracket)
                    level++;
                else if (text[i] == closeBracket)
                    level--;
            }
            return level == 0;
        }

        public static string[] GetUsingNamepseces(this SyntaxTree syntaxTree)
        {
            int iterator = 0;
            var nodes = new List<AstNode>(syntaxTree.Children);
            var usings = new List<string>();

            while (iterator < nodes.Count)
            {
                var node = nodes[iterator];
                nodes.AddRange(node.Children);

                var usingDecl = node as UsingDeclaration;
                if (usingDecl != null)
                    usings.Add(usingDecl.Namespace);

                iterator++;
            }

            return usings.ToArray();
        }

        public static string GetDisplayInfo(this ICompletionData data)
        {
            //TODO - needs to be revisited to implement non-EntityCompltiondata (e.g. enums) eventually
            var d = (data as EntityCompletionData);
            if (d != null)
            {
                return d.Entity.ToTooltip();
            }
            else
                return null;
        }

        public static IEnumerable<TSource> DistinctBy<TSource, TKey>(this IEnumerable<TSource> collection, Func<TSource, TKey> keySelector)
        {
            return collection.GroupBy(keySelector).Select(y => y.First());
        }
    }
}