﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Windows.Forms;

namespace CSScriptIntellisense
{
    public partial class MemberInfoPanel : Form, IPopupForm
    {
        public struct ItemInfo
        {
            public string Text;
            public int? ArgumentCount;

            public override string ToString()
            {
                return Text;
            }
        }

        int index = 0;

        public bool IsVisible()
        {
            return Visible;
        }

        int lastHintCount = -1;
        public void ProcessMethodOverloadHint(IEnumerable<string> hint)
        {
            int hitCount = (hint == null ? 0 : hint.Count());

            if (lastHintCount == hitCount)
                return; //nothing changed so no need to do anything

            lastHintCount = hitCount;

            string initaialSelection = (index == 0 ? null : items[index].Text);

            items.Clear();

            if (hitCount == 0)
                items.AddRange(rawItems);
            else
                items.AddRange(rawItems.Where(s => !s.ArgumentCount.HasValue || s.ArgumentCount.Value >= hitCount));

            index = 0;
            if (initaialSelection != null)
                for (int i = 0; i < items.Count; i++)
                    if (items[i].Text == initaialSelection)
                    {
                        index = i;
                        break;
                    }

            ResetActiveText();
        }

        public void AddData(IEnumerable<string> methodSignatures)
        {
            foreach (string item in methodSignatures)
            {
                ItemInfo info = new ItemInfo { Text = item };
                if (!Simple)
                    info.ArgumentCount = NRefactoryExtensions.GetArgumentCount(info.Text);
                items.Add(info);
            }

            rawItems.AddRange(items);

            index = 0;
            ResetActiveText();
        }

        string RemoveTypeCategory(string memberInfo)
        {
            return memberInfo.Split(new[] { ':' }, 2).Last();
        }

        void ResetActiveText()
        {
            SizeF size = MeasureDisplayArea();
            this.Width = (int)size.Width;
            this.Height = (int)size.Height;
            Invalidate();
        }

        public bool Simple = false;

        public bool AutoClose { get { return Simple; } }

        public MemberInfoPanel()
        {
            InitializeComponent();
        }

        public List<ItemInfo> items = new List<ItemInfo>();
        List<ItemInfo> rawItems = new List<ItemInfo>();

        public void kbdHook_KeyDown(Keys key, int repeatCount)
        {
            OnKeyDown(key);
        }

        void MemberInfoPanel_KeyDown(object sender, KeyEventArgs e)
        {
            OnKeyDown(e.KeyCode);
        }

        void OnKeyDown(Keys key)
        {
            if (Simple)
            {
                Close();
            }
            else
            {
                if (key == Keys.Escape || key == Keys.Enter)
                    Close();

                if (key == Keys.Up)
                    Decrement();

                if (key == Keys.Down && !Simple)
                    Increment();
            }
        }

        void QuickInfoPanel_Deactivate(object sender, EventArgs e)
        {
            if (Simple)
                Close();
        }

        void Increment()
        {
            index++;
            if (index >= items.Count)
                index = 0;
            ResetActiveText();
        }

        void Decrement()
        {
            index--;
            if (index < 0)
                index = items.Count - 1;
            ResetActiveText();
        }

        Brush bgdBrush = new SolidBrush(Color.FromArgb(0xE7E8EC));
        Pen borderPen = Pens.DarkGray;
        static float controlButtonSize = 8;
        float controlButtonsPadding = 3;

        SizeF MeasureDisplayArea()
        {
            string info = items[index].Text;

            if (!Simple)
                info = RemoveTypeCategory(info);

            SizeF size = this.CreateGraphics().MeasureString(info, this.Font);

            if (!Simple)
            {
                SizeF statsSize = this.CreateGraphics().MeasureString(GetOverloadingStats(), this.Font);
                size.Width = size.Width + statsSize.Width + controlButtonSize * 2 + controlButtonsPadding * 2 + 5;
            }

            return new SizeF(size.Width + 10, size.Height + 20);
        }

        RectangleF upButtonArea = new RectangleF(0, 0, controlButtonSize, controlButtonSize);
        RectangleF downButtonArea = new RectangleF(0, 0, controlButtonSize, controlButtonSize);

        void QuickInfoPanel_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            e.Graphics.PixelOffsetMode = PixelOffsetMode.Half;

            var area = this.ClientRectangle;

            float yOffset = 10;
            float xOffset = 5;

            e.Graphics.FillRectangle(bgdBrush, area);
            e.Graphics.DrawRectangle(borderPen, area);

            if (Simple)
            {
                e.Graphics.DrawString(items[index].Text, this.Font, Brushes.Black, xOffset, yOffset);
            }
            else
            {
                upButtonArea.X = xOffset;
                upButtonArea.Y = yOffset + 3f;
                e.Graphics.FillTranagle(Brushes.Black, xOffset, yOffset + 3f, controlButtonSize, controlButtonSize, true);

                xOffset += controlButtonSize + controlButtonsPadding;

                string stats = GetOverloadingStats();
                SizeF size = e.Graphics.MeasureString(stats, this.Font);
                e.Graphics.DrawString(stats, this.Font, Brushes.Black, xOffset, yOffset);
                xOffset += size.Width + controlButtonsPadding;

                downButtonArea.X = xOffset;
                downButtonArea.Y = yOffset + 3f;
                e.Graphics.FillTranagle(Brushes.Black, xOffset, yOffset + 3f, controlButtonSize, controlButtonSize, false);
                xOffset += controlButtonSize + controlButtonsPadding;

                e.Graphics.DrawString(RemoveTypeCategory(items[index].Text), this.Font, Brushes.Black, xOffset, yOffset);
            }
        }

        string GetOverloadingStats()
        {
            return string.Format("{0} of {1}", index + 1, items.Count);
        }

        private void MemberInfoPanel_MouseDown(object sender, MouseEventArgs e)
        {
            if (!Simple)
            {
                if (downButtonArea.Contains(e.Location))
                    Increment();
                else if (upButtonArea.Contains(e.Location))
                    Decrement();
            }
        }

        private void MemberInfoPanel_VisibleChanged(object sender, EventArgs e)
        {

        }

        private void MemberInfoPanel_Activated(object sender, EventArgs e)
        {
            var p = this.Parent;
        }
    }

    static class GraphicsExtensions
    {
        public static void FillTranagle(this Graphics g, Brush brush, float x, float y, float width, float height, bool pointUp)
        {
            var gp = new GraphicsPath(FillMode.Winding);

            if (pointUp)
                gp.AddLines(new PointF[] { new PointF(x, y + height), new PointF(x + width, y + height), new PointF(x + width / 2, y) });
            else
                gp.AddLines(new PointF[] { new PointF(x, y), new PointF(x + width, y), new PointF(x + width / 2, y + height) });

            g.FillPath(brush, gp);
        }
    }

    public interface IPopupForm
    {
        bool AutoClose { get; }
    }
}