using CSScriptIntellisense.Interop;
using ICSharpCode.NRefactory.Completion;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace CSScriptIntellisense
{
    static public partial class Plugin
    {
        public static string PluginName
        {
            get
            {
                if (Bootstrapper.IsInConflictWithCSScriptNpp())
                    return "C# Intellisence (disabled)";
                else
                    return "C# Intellisence";
            }
        }

        static internal string currentFile = null;
        static int currentFileCssHash = -1;
        static DateTime currentFileTimestamp;
        static List<string> parsedFiles = new List<string>();

        static bool SingleFileMode = false;
        static internal bool Enabled = false;

        static MemberInfoPopupManager memberInfoPopup;

        static internal void CommandMenuInit()
        {
            int cmdIndex = 0;
            CommandMenuInit(ref cmdIndex,
                (index, name, handler, isCtrl, isAlt, isShift, key) =>
                {
                    Plugin.SetCommand(index, name, handler, new ShortcutKey(isCtrl, isAlt, isShift, key));
                });
        }

        //this method will also be called from the parent plugin 
        static public void CommandMenuInit(ref int cmdIndex, CSScriptNpp.SetMenuCommand setCommand) 
        {
            //System.Diagnostics.Debug.Assert(false);

            bool standaloneSetup = (cmdIndex == 0);

            Plugin.Enabled = Bootstrapper.Init(standaloneSetup);

            if (Plugin.Enabled)
            {
                Dispatcher.Init();
                KeyInterceptor.Instance.Install();

                Task.Factory.StartNew(SimpleCodeCompletion.Init);

                setCommand(cmdIndex++, "Show auto-complete list", ShowSuggestionList, true, false, false, Keys.OemQuestion);
                setCommand(cmdIndex++, "Add missing 'using'", AddMissingUsings, true, false, false, Keys.OemPeriod);
                setCommand(cmdIndex++, "Re-analyze current document", Reparse, false, false, false, Keys.None);
                setCommand(cmdIndex++, "Show Method Info", ShowMethodInfo, false, false, false, Keys.F6);
                setCommand(cmdIndex++, "---", null, false, false, false, Keys.None);
                setCommand(cmdIndex++, "Settings", ShowConfig, false, false, false, Keys.None);
                setCommand(cmdIndex++, "---", null, false, false, false, Keys.None);
                if (standaloneSetup)
                    setCommand(cmdIndex++, "About", ShowAboutBox, false, false, false, Keys.None);

                memberInfoPopup = new MemberInfoPopupManager(ShowQuickInfo);
                KeyInterceptor.Instance.Add(Keys.Space);
                KeyInterceptor.Instance.KeyDown += Instance_KeyDown;
            }
            else
                SetCommand(cmdIndex++, "About", ShowAboutBox);
        }

        static void Instance_KeyDown(Keys key, int repeatCount, ref bool handled)
        {
            if (Config.Instance.InterceptCtrlSpace)
            {
                if (key == Keys.Space && KeyInterceptor.IsPressed(Keys.ControlKey) && Npp.IsCurrentFileHasExtension(".cs"))
                {
                    handled = true;
                    Dispatcher.Shedule(10, ShowSuggestionList);
                }
            }
        }

        static void ShowMethodInfo()
        {
            ShowInfo(false);
        }

        static void ShowQuickInfo()
        {
            ShowInfo(true);
        }

        static void ShowInfo(bool simple)
        {
            if (memberInfoPopup.IsShowing)
            {
                if (simple)
                    return;
                else
                    memberInfoPopup.Close();
            }

            string file = Npp.GetCurrentFile();

            if (Npp.IsCurrentFileHasExtension(".cs"))
            {
                int pos;

                if (simple)
                    pos = Npp.GetPositionFromMouseLocation();
                else
                    pos = Npp.GetCaretPosition();

                if (pos != -1)
                {
                    string text = Npp.TextBetween(0, Npp.DocEnd);

                    EnsureCurrentFileParsed();

                    int methodStartPos;

                    string[] data = SimpleCodeCompletion.GetMemberInfo(text, pos, file, simple, out methodStartPos);

                    if (data.Length > 0)
                    {
                        if (simple && Config.Instance.ShowQuickInfoInStatusBar)
                            Npp.SetStatusbarLabel(data.FirstOrDefault() ?? defaultStatusbarLabel);
                        else
                            memberInfoPopup.TriggerPopup(simple, methodStartPos, data);
                    }
                }
            }
        }

        static CustomContextMenu namespaceMenu;

        static void AddMissingUsings()
        {
            if (Npp.IsCurrentFileHasExtension(".cs"))
            {
                IEnumerable<TypeInfo> items = ResolveNamespacesAtCaret().ToArray();

                if (items.Count() > 0)
                {
                    Point point = Npp.GetCaretScreenLocation();

                    if (namespaceMenu != null && namespaceMenu.Visible)
                        namespaceMenu.Close();

                    namespaceMenu = new CustomContextMenu();
                    namespaceMenu.Left = point.X;
                    namespaceMenu.Top = point.Y + 18;

                    var usings = items.Where(x => !x.IsNested)
                                      .Select(x => "using " + x.Namespace + ";");

                    var inline = items.Select(x => x.FullName);

                    foreach (string item in usings)
                        namespaceMenu.Add(item, Images.Images.namespace_add, NppEditor.InsertNamespace);

                    namespaceMenu.AddSeparator();

                    foreach (string item in inline)
                        namespaceMenu.Add(item, null, Npp.ReplaceWordAtCaret);

                    namespaceMenu.Popup();
                }
            }
        }

        static void ShowAboutBox()
        {
            using (var form = new AboutBox())
                form.ShowDialog();
        }

        static void ShowConfig()
        {
            using (var form = new ConfigForm(Config.Instance))
            {
                form.ShowDialog();
                Config.Instance.Save();
            }
        }

        static AutocompleteForm form;

        static void ShowSuggestionList()
        {
            if (Npp.IsCurrentFileHasExtension(".cs"))
            {
                var items = GetSuggestionItemsAtCaret();

                if (items.Count() > 0)
                {
                    bool memberInfoWasShowing = memberInfoPopup.IsShowing;
                    if (memberInfoWasShowing)
                        memberInfoPopup.Close();

                    Point point = Npp.GetCaretScreenLocation();

                    if (form != null && form.Visible)
                        form.Close();

                    form = new AutocompleteForm(OnAutocompletionAccepted, items, NppEditor.GetSuggestionHint());
                    form.Left = point.X;
                    form.Top = point.Y + 18;
                    form.FormClosed += (sender, e) =>
                                        {
                                            if (memberInfoWasShowing)
                                                ShowMethodInfo();
                                        };
                    form.KeyPress += (sender, e) =>
                                        {
                                            if (e.KeyChar >= ' ' || e.KeyChar == 8) //8 is backspace
                                                OnAutocompleteKeyPress(e.KeyChar);
                                        };
                    form.Show();

                    OnAutocompleteKeyPress();
                }
            }
            else if (Config.Instance.InterceptCtrlSpace)
            {
                const int WM_COMMAND = 0x0111;
                const int MENU_FUNCTION_COMPLETE = 50000;
                Win32.SendMessage(Plugin.NppData._nppHandle, (NppMsg)WM_COMMAND, MENU_FUNCTION_COMPLETE, 0);
            }
        }

        static void OnAutocompletionAccepted(ICompletionData data)
        {
            if (data != null)
            {
                IntPtr sci = GetCurrentScintilla();

                int currentPos = Npp.GetCaretPosition();
                Point p;
                string word = Npp.GetWordAtCursor(out p);

                if (word != "")  // e.g. Console.Wr| but not Console.|
                {
                    if (!word.EndsWith("->"))
                    {
                        Win32.SendMessage(sci, SciMsg.SCI_SETSELECTION, p.X, p.Y);
                    }
                }

                Win32.SendMessage(sci, SciMsg.SCI_REPLACESEL, 0, data.CompletionText);
            }

            if (form != null)
            {
                if (form.Visible)
                    form.Close();
                form = null;
            }
        }

        static void OnAutocompleteKeyPress(char keyChar = char.MinValue)
        {
            string justTypedText = NppEditor.ProcessKeyPress(keyChar);

            int currentPos = Npp.GetCaretPosition();

            string text = Npp.TextBetween(Math.Max(0, currentPos - 30), currentPos); //check up to 30 chars from left
            int pos = text.LastIndexOfAny(SimpleCodeCompletion.Delimiters);
            if (pos != -1)
            {
                string token = text.Substring(pos + 1) + justTypedText;
                form.FilterFor(token.Trim());
            }
            else
            {
                form.Close();
            }
        }

        static void Reparse()
        {
            parsedFiles.Clear();
            currentFile = null;
            EnsureCurrentFileParsedAsynch();
        }

        static public void OnCharTyped()
        {
            if (Npp.IsCurrentFileHasExtension(".cs"))
            {
                string text = Npp.TextBeforeCursor(1);
                string text2 = Npp.TextBeforeCursor(2);
                if (text == "." || text == "&" || text2 == "->")
                    ShowSuggestionList();
                else if (text == "(")
                {
                    if (!memberInfoPopup.IsShowing || memberInfoPopup.Simple)
                        ShowMethodInfo();
                }
                else if (memberInfoPopup.IsShowing)
                    memberInfoPopup.CheckIfNeedsClosing();
            }
        }

        static string defaultStatusbarLabel = "C# source file";

        static public void OnCurrentFileChanegd()
        {
            if (Plugin.Enabled)
            {
                Plugin.EnsureCurrentFileParsedAsynch();

                if (Npp.IsCurrentFileHasExtension(".cs"))
                    memberInfoPopup.Enabled = true;
                else
                    memberInfoPopup.Enabled = false;
            }
        }

        static public void OnNppReady()
        {
            Plugin.EnsureCurrentFileParsedAsynch();
        }

        static public void EnsureCurrentFileParsedAsynch()
        {
            if (Plugin.Enabled)
                Task.Factory.StartNew(EnsureCurrentFileParsed);
        }

        static public Func<string> ResolveCurrentFile = Npp.GetCurrentFile; //the implementation can be injected by the host or other plugins. To be used in the future.

        static public void EnsureCurrentFileParsed()
        {
            if (Plugin.Enabled)
                lock (typeof(Plugin))
                {
                    string file = ResolveCurrentFile();

                    if (string.IsNullOrWhiteSpace(file) || !file.EndsWith(".cs", StringComparison.OrdinalIgnoreCase))
                        return;

                    int newCssHash = -1;
                    //when to regenerate the project:
                    // - not initialized yet
                    // - new file which is not a part of the 'current' project
                    // - the current file changed and saved with the new set of CS-Script instructions (for multi-file mode only)
                    if (currentFile == null ||
                       (currentFile != file && !parsedFiles.Contains(file)) ||
                       (!SingleFileMode && currentFileTimestamp != File.GetLastWriteTime(file) && currentFileCssHash != (newCssHash = NppEditor.GetCssHash(file))))
                    {
                        currentFile = file;
                        currentFileCssHash = (newCssHash != -1) ? newCssHash : NppEditor.GetCssHash(file);
                        currentFileTimestamp = File.GetLastWriteTime(file);

                        Tuple<string[], string[]> project = CSSctiptHelper.GetProjectFiles(file);

                        string[] sourceFiles = project.Item1;
                        string[] assemblyFiles = project.Item2;

                        var sourcesInfos = new List<Tuple<string, string>>();

                        if (SingleFileMode)
                        {
                            sourcesInfos.Add(new Tuple<string, string>(Npp.TextBetween(0, Npp.DocEnd), file));
                        }
                        else
                        {
                            foreach (string item in sourceFiles)
                            {
                                sourcesInfos.Add(new Tuple<string, string>(File.ReadAllText(item), item));
                            }
                        }

                        SimpleCodeCompletion.ResetProject(sourcesInfos.ToArray(), assemblyFiles);
                    }
                }
        }

        static IEnumerable<ICompletionData> GetSuggestionItemsAtCaret()
        {
            string file = Npp.GetCurrentFile();
            string text = Npp.TextBetween(0, Npp.DocEnd);
            int currentPos = Npp.GetCaretPosition();

            EnsureCurrentFileParsed();
            return SimpleCodeCompletion.GetCompletionData(text, currentPos, Npp.GetCurrentFile()).Where(item => item != null);
        }

        static IEnumerable<TypeInfo> ResolveNamespacesAtCaret()
        {
            string file = Npp.GetCurrentFile();
            string text = Npp.TextBetween(0, Npp.DocEnd);
            int currentPos = Npp.GetCaretPosition();

            EnsureCurrentFileParsed();
            return SimpleCodeCompletion.GetMissingUsings(text, currentPos, Npp.GetCurrentFile());
        }
    }
}